library ieee;
use ieee.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.defs.all;

entity enc_top is
	port(
		clk       : in  std_logic;
		rst       : in  std_logic;
		data_in   : in  std_logic_vector(7 downto 0);
		new_frame : in  std_logic;
		ready     : out std_logic;
		data_out  : out std_logic_vector(GOL_MAX_LEN - 1 downto 0);
		size_out  : out std_logic_vector(3 downto 0)
	);
end entity enc_top;

architecture RTL of enc_top is
	signal data_in_c : std_logic_vector(7 downto 0);

	--index
	signal index_rst : std_logic;
	signal B_i       : std_logic_vector(SMPL_LEN - 1 downto 0);
	signal C_i       : std_logic_vector(SMPL_LEN - 1 downto 0);
	signal D_i       : std_logic_vector(SMPL_LEN - 1 downto 0);
	signal I_i       : std_logic_vector(SMPL_LEN - 1 downto 0);
	signal index_out : std_logic_vector(CTXT_ADDR_SIZE - 1 downto 0);
	signal sign_out  : std_logic;

	--k
	signal loco_k_rst : std_logic;

	signal A_k_out    : std_logic_vector(A_LEN - 1 downto 0);
	signal N_k_out    : std_logic_vector(N_LEN - 1 downto 0);
	signal k_out      : std_logic_vector(8 downto 0);
	signal k_less_out : std_logic_vector(8 downto 0);
	signal k_int_out  : std_logic_vector(3 downto 0);

	--loco
	signal loco_rst : std_logic;

	signal A_out      : std_logic_vector(A_LEN - 1 downto 0);
	signal B_out      : std_logic_vector(B_LEN - 1 downto 0);
	signal C_out      : std_logic_vector(C_LEN - 1 downto 0);
	signal N_out      : std_logic_vector(N_LEN - 1 downto 0);
	signal B_l        : std_logic_vector(SMPL_LEN - 1 downto 0);
	signal sign_d     : std_logic;
	signal loco_w_rst : std_logic;
	signal index_d2   : std_logic_vector(7 downto 0);
	signal A_in       : std_logic_vector(A_LEN - 1 downto 0);
	signal B_in       : std_logic_vector(B_LEN - 1 downto 0);
	signal C_in       : std_logic_vector(C_LEN - 1 downto 0);
	signal N_in       : std_logic_vector(N_LEN - 1 downto 0);
	signal Q          : std_logic_vector(SMPL_LEN - 1 downto 0);

	--gol
	signal gol_rst  : std_logic;
	signal k_less_d : std_logic_vector(8 downto 0);
	signal k_d      : std_logic_vector(8 downto 0);
	signal k_int_d  : std_logic_vector(3 downto 0);

	--context
	signal ctxt_addr_rst : std_logic_vector(CTXT_ADDR_SIZE - 1 downto 0);
	signal ctxt_rst      : std_logic;

	--output
	signal data_out_c : std_logic_vector(GOL_MAX_LEN - 1 downto 0);
	signal size_out_c : std_logic_vector(3 downto 0);

	signal cache_rst  : std_logic;
	signal not_rst    : std_logic;
	signal index_d1   : std_logic_vector(7 downto 0);
	signal B_gol      : std_logic_vector(B_LEN - 1 downto 0);
	signal N_gol      : std_logic_vector(N_LEN - 1 downto 0);
	signal data_in_c1 : std_logic_vector(7 downto 0);
	signal data_in_c2 : std_logic_vector(7 downto 0);
	signal data_in_c3 : std_logic_vector(7 downto 0);
begin
	not_rst <= not rst;

	clocked_in_out : process(clk, rst) is
	begin
		if rising_edge(clk) then
			if rst = '1' then
				data_in_c <= (others => '0');
				data_out  <= (others => '0');
				size_out  <= (others => '0');
			else
				data_in_c  <= data_in;
				data_in_c3 <= data_in_c2;
				data_in_c2 <= data_in_c1;
				data_in_c1 <= data_in_c;

				data_out <= data_out_c;
				size_out <= size_out_c;
			end if;
		end if;
	end process clocked_in_out;

	index_module : entity work.index port map(
			clk       => clk,
			rst       => index_rst,
			B_cache   => B_i,
			C_cache   => C_i,
			D_cache   => D_i,
			I_cache   => I_i,
			index_out => index_out,
			sign_out  => sign_out
		);
	find_k_module : entity work.find_k port map(
			clk        => clk,
			rst        => loco_k_rst,
			A          => A_k_out,
			N          => N_k_out,
			k_out      => k_out,
			k_less_out => k_less_out,
			k_int_out  => k_int_out
		);
	modeling_module : entity work.loco port map(
			clk        => clk,
			rst        => loco_rst,
			loco_w_rst => loco_w_rst,
			data       => data_in_c,
			sign       => sign_d,
			P          => B_l,
			A          => A_out,
			B          => B_out,
			C          => C_out,
			N          => N_out,
			A_out      => A_in,
			B_out      => B_in,
			C_out      => C_in,
			N_out      => N_in,
			B0_out     => B_gol,
			N0_out     => N_gol,
			Q_out      => Q
		);
	gol_module : entity work.gol_enc_top port map(
			clk        => clk,
			rst        => gol_rst,
			k_in       => k_d,
			k_less_in  => k_less_d,
			k_int_in   => k_int_d,
			input      => Q,
			B          => B_gol,
			N          => N_gol,
			output     => data_out_c,
			length_out => size_out_c
		);
	ctxt_reg_module : entity work.ctxt_reg port map(
			clk          => clk,
			rst          => rst,
			write_0      => ctxt_rst,
			write0_addr  => ctxt_addr_rst,
			read_index   => index_d1,
			read_k_index => index_out,
			write_index  => index_d2,
			A_in         => A_in,
			B_in         => B_in,
			C_in         => C_in,
			N_in         => N_in,
			A_out        => A_out,
			A_k_out      => A_k_out,
			B_out        => B_out,
			C_out        => C_out,
			N_out        => N_out,
			N_k_out      => N_k_out
		);
	cache_module : entity work.line_cache port map(
			clk        => clk,
			shift      => not_rst,
			write_0    => cache_rst,
			sr_in      => data_in_c3,
			sr_tap_B   => B_i,
			sr_tap_C   => C_i,
			sr_tap_D   => D_i,
			sr_tap_I   => I_i,
			sr_tap_B_l => B_l
		);
	delay_module : entity work.enc_signal_delays port map(
			clk        => clk,
			rst        => rst,
			index_in   => index_out,
			sign_in    => sign_out,
			k_less_in  => k_less_out,
			k_in       => k_out,
			k_int_in   => k_int_out,
			index_out1 => index_d1,
			index_out2 => index_d2,
			sign_out   => sign_d,
			k_less_out => k_less_d,
			k_out      => k_d,
			k_int_out  => k_int_d
		);
	reset_module : entity work.reset port map(
			clk        => clk,
			rst        => rst,
			new_frame  => new_frame,
			ready      => ready,
			ctxt_addr  => ctxt_addr_rst,
			ctxt_rst   => ctxt_rst,
			cache_rst  => cache_rst,
			index_rst  => index_rst,
			loco_k_rst => loco_k_rst,
			loco_rst   => loco_rst,
			loco_w_rst => loco_w_rst,
			gol_rst    => gol_rst
		);
end architecture RTL;
