library ieee;
use ieee.std_logic_1164.all;
use ieee.STD_LOGIC_ARITH.ALL;
use ieee.std_logic_unsigned.all;
use work.defs.all;

entity reset is
	port(
		clk        : in  std_logic;
		rst        : in  std_logic;
		new_frame  : in  std_logic;
		ready      : out std_logic;
		ctxt_addr  : out std_logic_vector(CTXT_ADDR_SIZE - 1 downto 0);
		ctxt_rst   : out std_logic;
		cache_rst  : out std_logic;
		index_rst  : out std_logic;
		loco_k_rst : out std_logic;
		loco_rst   : out std_logic;
		loco_w_rst : out std_logic;
		gol_rst    : out std_logic
	);
end entity reset;

architecture RTL of reset is
	signal counter     : std_logic_vector(12 downto 0);
	signal index_rst0  : std_logic_vector(ENC_INDEX_RST downto 0);
	signal loco_k_rst0 : std_logic_vector(ENC_K_RST downto 0);
	signal loco_rst0   : std_logic_vector(ENC_LOCO_RST downto 0);
	signal gol_rst0    : std_logic_vector(ENC_GOL_RST downto 0);
	signal loco_w_rst0 : std_logic_vector(ENC_LOCO_W_RST downto 0);
begin

	-- Count through all addresses in context memory and write 0:
	empty_cache : process(clk) is
	begin
		if rising_edge(clk) then
			if rst = '1' then
				counter   <= (others => '0');
				ctxt_addr <= (others => '0');
			else
				if new_frame = '1' then
					counter   <= counter + 1;
					ctxt_addr <= counter(CTXT_ADDR_SIZE - 1 downto 0);
				else
					counter   <= (others => '0');
					ctxt_addr <= (others => '0');
				end if;
			end if;
		end if;
	end process empty_cache;

	-- Delay reset signals to all modules in the design:
	reset : process(clk) is
	begin
		if rising_edge(clk) then
			if rst = '1' then
				ready       <= '0';
				index_rst   <= '1';
				index_rst   <= '1';
				loco_k_rst  <= '1';
				loco_rst    <= '1';
				loco_w_rst  <= '1';
				gol_rst     <= '1';
				index_rst0  <= (others => '1');
				loco_k_rst0 <= (others => '1');
				loco_rst0   <= (others => '1');
				gol_rst0    <= (others => '1');
			else
				if new_frame = '1' then
					-- After counting two lines and line cache is empty:
					if counter(12) = '1' then
						-- Set input value:
						ready                                <= '1';
						index_rst0(0)                        <= '0';
						loco_k_rst0(0)                       <= '0';
						loco_rst0(0)                         <= '0';
						loco_w_rst0(0)                       <= '0';
						gol_rst0(0)                          <= '0';
						
						-- Propagate signals:
						index_rst0(ENC_INDEX_RST downto 1)   
							<= index_rst0(ENC_INDEX_RST - 1 downto 0);
						loco_k_rst0(ENC_K_RST downto 1)      
							<= loco_k_rst0(ENC_K_RST - 1 downto 0);
						loco_rst0(ENC_LOCO_RST downto 1)     
							<= loco_rst0(ENC_LOCO_RST - 1 downto 0);
						loco_w_rst0(ENC_LOCO_W_RST downto 1) 
							<= loco_w_rst0(ENC_LOCO_W_RST - 1 downto 0);
						gol_rst0(ENC_GOL_RST downto 1)       
							<= gol_rst0(ENC_GOL_RST - 1 downto 0);
					else
						-- Else keep reset:
						ready       <= '0';
						index_rst0  <= (others => '1');
						loco_k_rst0 <= (others => '1');
						loco_rst0   <= (others => '1');
						loco_w_rst0 <= (others => '1');
						gol_rst0    <= (others => '1');
					end if;
				else
					-- Keep not reset after new frame has started:
					ready                                <= '0';
					index_rst0(0)                        <= '0';
					loco_k_rst0(0)                       <= '0';
					loco_rst0(0)                         <= '0';
					loco_w_rst0(0)                       <= '0';
					gol_rst0(0)                          <= '0';
					index_rst0(ENC_INDEX_RST downto 1)   
						<= index_rst0(ENC_INDEX_RST - 1 downto 0);
					loco_k_rst0(ENC_K_RST downto 1)      
						<= loco_k_rst0(ENC_K_RST - 1 downto 0);
					loco_rst0(ENC_LOCO_RST downto 1)     
						<= loco_rst0(ENC_LOCO_RST - 1 downto 0);
					loco_w_rst0(ENC_LOCO_W_RST downto 1) 
						<= loco_w_rst0(ENC_LOCO_W_RST - 1 downto 0);
					gol_rst0(ENC_GOL_RST downto 1)       
						<= gol_rst0(ENC_GOL_RST - 1 downto 0);
				end if;
				-- Output delayed reset signals:
				index_rst  <= index_rst0(ENC_INDEX_RST);
				loco_k_rst <= loco_k_rst0(ENC_K_RST);
				loco_rst   <= loco_rst0(ENC_LOCO_RST);
				loco_w_rst <= loco_w_rst0(ENC_LOCO_W_RST);
				gol_rst    <= gol_rst0(ENC_GOL_RST);
				ctxt_rst   <= index_rst0(ENC_INDEX_RST);
				cache_rst  <= index_rst0(ENC_INDEX_RST);
			end if;
		end if;
	end process reset;
end architecture RTL;