library ieee;
use ieee.std_logic_1164.all;
use ieee.STD_LOGIC_ARITH.ALL;
use ieee.std_logic_unsigned.all;
use work.defs.all;

--
-- Golomb Packer (WRITE BUFFER)
-- Collect encoded words from golomb encoder (genc) 
-- and write Golomb packages when they are ready

entity wbuf is
	generic(                            -- number of input bits
		constant IN_WIDTH   : integer := 16;
		-- number of output bits		
		constant OUT_WIDTH  : integer := 16;
		-- number of bits in the internal buffer					
		constant WBUF_WIDTH : integer := 32
	);

	port(
		clk         : in  std_ulogic;
		rst         : in  std_ulogic;

		-- INPUTS
		-- Input from the golomb encoder
		in_wbuf     : in  std_logic_vector(IN_WIDTH - 1 downto 0);
		-- Length of the valid part of the input
		word_length : in  std_logic_vector(3 downto 0);
		-- Enable the wbuf Golomb packer			
		wbuf_enable : in  std_logic;

		-- OUTPUTS
		-- Golomb packer output
		out_wbuf    : out std_logic_vector(OUT_WIDTH - 1 downto 0);
		-- The golomb packer is empty, 
		-- do not expect additional Golomb packages
		wbuf_empty  : out std_logic;
		-- A Golomb packet is ready on the output								
		wbuf_write  : out std_logic
	);

end entity wbuf;

architecture RTL of wbuf is
begin
	G_ENC : process(clk, rst, wbuf_enable) is
		variable wbuffer   : std_logic_vector(WBUF_WIDTH - 1 downto 0);
		variable init_done : boolean;

		variable num_elements : std_logic_vector(4 downto 0);
		variable write_pos    : std_logic_vector(4 downto 0);
		variable read_pos     : std_logic_vector(4 downto 0);

		--variable shiftL			: std_ulogic_vector(4 downto 0);

		variable shift_input : std_logic_vector(in_wbuf'range);
	begin
		if (rst = '1') then
			wbuffer      := (others => '0');
			init_done    := false;
			num_elements := (others => '0');
			write_pos    := (others => '1');
			read_pos     := (others => '1');

			out_wbuf   <= (others => '0');
			wbuf_empty <= '0';
			wbuf_write <= '0';

		elsif (rising_edge(clk) and wbuf_enable = '1') then
			if (init_done) then

				-- Shifting the input to the left so that the valid part 
				-- starts at position 15 of 'shift_input'.
				-- This is done because the input (all 16 bits) can then 
				-- be written directly to the array.
				-- The alternative would be to pick a variable number of 
				-- bits and place them into the array, which
				-- is expected to increase the module complexity\size.
				shift_input(
					IN_WIDTH - 1 downto IN_WIDTH
					- conv_integer((word_length)) - 1
				) := in_wbuf(conv_integer((word_length)) downto 0);

				-- Checks to see if the 'write_pos' is lower than 16, 
				-- which will make the input wrap around 'wbuffer' 
				-- and continue at the other end.
				-- (else copies 'shift_input' straight into 'wbuffer'
				if (conv_integer(write_pos) < IN_WIDTH) then
					wbuffer(
						conv_integer(write_pos) downto 0
					) := shift_input(
						IN_WIDTH - 1 downto IN_WIDTH - 1 - conv_integer(write_pos)
					);

					wbuffer(
						WBUF_WIDTH - 1 downto WBUF_WIDTH - 1
						  - (IN_WIDTH - 1 - conv_integer(write_pos) - 1)
					) := shift_input(
						IN_WIDTH - 1 - conv_integer(write_pos) - 1 downto 0
					);
				else
					wbuffer(
						conv_integer(write_pos) 
						downto conv_integer(write_pos) - IN_WIDTH + 1
					) := shift_input;
				end if;

				-- Add 'word_length' to the number of elements (bits) 
				-- currently in the wbuffer
				num_elements := num_elements + (word_length) + "1";
				-- Move the write position by subtracting the word_length 
				-- (word_length from 0 to 15 represents a code word 
				-- length of 1 to 16 bits) 
				write_pos    := write_pos - (word_length) - "1";

				-- Writes data to the output if the number of 
				-- elements is more than the Output width.
				-- wbuf_write is set to 1 to indicate that a 
				-- Golomb package was written
				-- (else sets wbuf_write back to 0)
				if (num_elements >= OUT_WIDTH) then
					out_wbuf     <= wbuffer(
														conv_integer(read_pos) 
														downto conv_integer(read_pos)
														- OUT_WIDTH + 1
													);
					wbuf_write   <= '1';
					read_pos     := read_pos - "10000";
					num_elements := num_elements - "10000";
				else
					wbuf_write <= '0';
					out_wbuf   <= (others => '0');
				end if;
				wbuf_empty <= '0';
			else
				num_elements := (others => '0');
				wbuf_empty   <= '1';
				write_pos    := (others => '1');
				read_pos     := (others => '1');
				init_done    := true;
			end if;

		-- When the enable goes low, there could still be data left 
		-- in the write buffer. This ensures that the last golomb 
		-- package is written even if it doesnt contain only valid bits.
		-- A counter in the decoder will know when all pixels have been 
		-- decoded, and the unvalid bits will be ignored
		elsif (rising_edge(clk) and wbuf_enable = '0') then
			if (num_elements > 0) then
				out_wbuf     <= wbuffer(
													conv_integer(read_pos) 
													  downto conv_integer(read_pos)
													  - OUT_WIDTH + 1
												);
				wbuf_write   <= '1';
				wbuf_empty   <= '0';
				num_elements := (others => '0');
			else
				wbuf_empty <= '1';
				wbuf_write <= '0';
				out_wbuf   <= (others => '0');
				read_pos   := (others => '1');
				write_pos  := (others => '1');
				init_done  := false;
			end if;
		end if;
	end process;
end architecture RTL;
