library ieee;
library modelsim_lib;
use ieee.std_logic_1164.all;
use IEEE.STD_LOGIC_ARITH.ALL;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use modelsim_lib.util.all;
use work.defs.all;

-- Decoder testbench:
-- Instantiates 3 decoders (R,G,B)
-- Reads inputs from 3 (proprietary) input files, 
-- created by the Encoder testbench
-- Produces 1 output file, containing uncompressed data, 
-- identical to the Encoder input file
-- No statistics are calculated. 
-- 
-- NOTE: Input files read from 'path\fil'  ,  output file written to  
-- 'path\', where 'path' and 'fil' are defined below.
-- 'fil' folder must be created manually within the 'path' folder. 
-- The simulation cannot create folders. 

entity dec_tb is
	generic(
		constant IN_WIDTH  : integer := 16;
		constant OUT_WIDTH : integer := 8;
		-- Half the clock period. 
		-- The frequency will be 1/(2*PERIOD) = 100 MHz
		constant PERIOD    : time    := 50 ns;
		-- file name affix ( Note: 	the simulation expects the file
		--                   	 	to have 'vhdl' as an suffix)
		constant fil       : string  := "input_file";

		-- Folder where input file is located
		constant path      : string  := "..\source\";
		-- Number of pixels in frame data
		constant size      : integer := 1920 * 1080 
	);
end entity dec_tb;

architecture RTL of dec_tb is
	signal clk    : std_logic;
	signal rst    : std_logic;
	signal inputR : std_logic_vector(IN_WIDTH - 1 downto 0);
	signal inputG : std_logic_vector(IN_WIDTH - 1 downto 0);
	signal inputB : std_logic_vector(IN_WIDTH - 1 downto 0);

	signal outputR : std_logic_vector(OUT_WIDTH - 1 downto 0);
	signal outputG : std_logic_vector(OUT_WIDTH - 1 downto 0);
	signal outputB : std_logic_vector(OUT_WIDTH - 1 downto 0);

	signal readR, readG, readB : std_logic;
	signal rdyR, rdyG, rdyB    : std_logic;
	signal newFrame            : std_logic;

	signal active : std_logic := '0';

	shared variable activate            : std_logic := '0';
	shared variable numPixelsIn         : integer   := 0;
	shared variable numPixelsOut        : integer   := 1;
	shared variable bitsR, bitsG, bitsB : real      := 0.0;

	-- Function that reverses a std_logic_vector
	function reverse_any_vector(a : in std_logic_vector) 
		return std_logic_vector is
		variable result : std_logic_vector(a'RANGE);
		alias aa        : std_logic_vector(a'REVERSE_RANGE) is a;
	begin
		for i in aa'RANGE loop
			result(i) := aa(i);
		end loop;
		return result;
	end;

begin

	-- DECODER INIT
	decR : entity work.dec_top
		port map(
			clk       => clk,
			rst       => rst,
			data_in   => inputR,
			new_frame => newFrame,
			ready     => rdyR,
			ready_gol => readR,
			data_out  => outputR
		);

	decG : entity work.dec_top
		port map(
			clk       => clk,
			rst       => rst,
			data_in   => inputG,
			new_frame => newFrame,
			ready     => rdyG,
			ready_gol => readG,
			data_out  => outputG
		);

	decB : entity work.dec_top
		port map(
			clk       => clk,
			rst       => rst,
			data_in   => inputB,
			new_frame => newFrame,
			ready     => rdyB,
			ready_gol => readB,
			data_out  => outputB
		);

	--/ ---------------------------------
	--/ GENERATES THE CLOCK
	--/
	clock_generator : process
	begin
		clk <= '1';
		wait for PERIOD / 2;
		clk <= '0';
		wait for PERIOD / 2;
	end process;

	--/ ---------------------------------
	--/ CONNECT SIGNALS ACROSS DIFFERENT 
	--/ MODULES
	--/ Used to probe entity signals that aren't 
	--/ connected directly to ports
	--drive_sig_process : process
	--  begin
	--    init_signal_driver("/dec_tb/encR/Q", "qR");
	--    init_signal_driver("/dec_tb/encG/Q", "qG");
	--	init_signal_driver("/dec_tb/encB/Q", "qB");    
	--	wait;
	--end process drive_sig_process;
	-- ---------------------------------


	--/ ---------------------------------
	--/  ACTIVATE OUTPUT WRITING
	--/  Sets the active signal high at the
	--/  right time after activate has been  
	--/  set in the input process below
	act : process(clk)
		variable count : integer := 0;
	begin
		if (rising_edge(clk)) then
			if (activate = '1') then
				if (count > 13) then
					active <= '1';
				else
					count := count + 1;
				end if;
			else
				active <= '0';
			end if;
		end if;
	end process;

	-- ---------------------------------
	--/ PROCESS CONTROLLING INPUT
	--/ Input is read from the files defined with the constant 'fil'
	--/ Waits until decoder module is ready, then sets input,
	--/ listens to 'read' port, updates inputs when requested.
	--/ Loops as long as at least one input file is not empty
	--/ Leaves loop when files are empty, waits for Output process
	--/ to finish and end the simulation.

	inputctrl : process
		file infileR : text is in path & fil & "\" & fil & "_outR";
		file infileG : text is in path & fil & "\" & fil & "_outG";
		file infileB : text is in path & fil & "\" & fil & "_outB";

		file readorder : text open WRITE_MODE 
		  is path & fil & "\" & fil & "_readorder";

		variable buf_inR, buf_inG, buf_inB, buf_nSymb, buf_rOrdr 
		: line;
		variable inR, inG, inB                                   
		: std_logic_vector(IN_WIDTH - 1 downto 0) := (others => '0');
		variable readVector                                      
		: std_logic_vector(2 downto 0)            := "000";
	begin
		rst <= '1';
		wait for PERIOD;
		inputR <= (others => '0');
		inputG <= (others => '0');
		inputB <= (others => '0');
		rst    <= '0';

		wait for 5 * PERIOD;
		newFrame <= '1';
		READLINE(infileR, buf_inR);
		READ(buf_inR, inR);
		readVector(2) := '1';

		READLINE(infileG, buf_inG);
		READ(buf_inG, inG);
		readVector(1) := '1';

		READLINE(infileB, buf_inB);
		READ(buf_inB, inB);
		readVector(0) := '1';

		inR := reverse_any_vector(inR);
		inG := reverse_any_vector(inG);
		inB := reverse_any_vector(inB);

		inputR <= inR;
		inputG <= inG;
		inputB <= inB;
		while (rdyR = '0' or rdyG = '0' or rdyB = '0') loop
			wait for PERIOD;
		end loop;
		wait for 5 ns;
		newFrame <= '0';
		wait for 45 ns;
		activate := '1';

		-- LOOP AS LONG AS AT LEAST ONE FILE IS NOT EMPTY
		while (not ENDFILE(infileR)) or (not ENDFILE(infileG)) 
		or (not ENDFILE(infileB)) loop
			readVector := "000";
			if (readR = '1' and not ENDFILE(infileR)) then
				READLINE(infileR, buf_inR);
				READ(buf_inR, inR);
				inR := reverse_any_vector(inR);

				readVector(2) := '1';
			end if;

			if (readG = '1' and not ENDFILE(infileG)) then
				READLINE(infileG, buf_inG);
				READ(buf_inG, inG);
				inG := reverse_any_vector(inG);

				readVector(1) := '1';
			end if;

			if (readB = '1' and not ENDFILE(infileB)) then
				READLINE(infileB, buf_inB);
				READ(buf_inB, inB);
				inB := reverse_any_vector(inB);

				readVector(0) := '1';
			end if;

			WRITE(buf_rOrdr, conv_integer(readVector(2)));
			WRITE(buf_rOrdr, string'(" "));
			WRITE(buf_rOrdr, conv_integer(readVector(1)));
			WRITE(buf_rOrdr, string'(" "));
			WRITE(buf_rOrdr, conv_integer(readVector(0)));
			WRITELINE(readorder, buf_rOrdr);

			-- Set input signals			
			wait for 5 ns;
			inputR <= inR;
			inputG <= inG;
			inputB <= inB;
			-- Count pixel
			wait for 45 ns;
		--			wait for PERIOD;
		end loop;

		wait for 8 * PERIOD;
		wait;
	end process;

	-- ---------------------------------
	--/ PROCESS CONTROLLING OUTPUT
	--/ Output starts writing when active is high.
	--/	Continues until the correct number of pixels is reached,
	--/ as defined by the 'size' constant.
	--/ Compares outputs with the values in the original input file.
	--/ 

	writectrl : process(clk)
		file originalFile : text open READ_MODE is path & fil & "vhdl";
		file outFile : text open WRITE_MODE is path & fil & "_out";

		variable buf_inOriginal, buf_out : line;
		variable decoded, original : std_logic_vector(24 - 1 downto 0);

	begin
		if (rising_edge(clk)) then
			if (active = '1' and numPixelsOut < (size) + 1) then

				-- Spectral "recorrelation"
				decoded := (outputR + outputG)&outputG&(outputB + outputG);

				-- DEBUG UTEN R-G,G,B-G:
				--decoded := outputR & outputG & outputB;

				HWRITE(buf_out, decoded, RIGHT, 6);
				WRITELINE(outFile, buf_out);
				numPixelsOut := numPixelsOut + 1;
				if (not ENDFILE(originalFile)) then
					READLINE(originalFile, buf_inOriginal);
					HREAD(buf_inOriginal, original);
					if (original /= decoded) then
						WRITE(OUTPUT, "Output did not match the original value ");
						WRITE(
							OUTPUT, "on pixel number " 
							& natural'image(numPixelsOut) & "." & LF
						);
					end if;
				end if;
			elsif (numPixelsOut > size) then
				-- -------------------------------------
				-- END SIMULATION : 
				-- -------------------------------------
				write(OUTPUT, " ----------------- " & LF);
				write(OUTPUT, "Simulation complete.");
				report "" severity failure;
			-- -------------------------------------
			else
			-- something , waiting for active.
			end if;
		end if;
	end process;
end architecture RTL;
