library ieee;
library modelsim_lib;
use ieee.std_logic_1164.all;
use IEEE.STD_LOGIC_ARITH.ALL;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use modelsim_lib.util.all;
use work.defs.all;

-- Encoder testbench:
-- Instantiates 3 encoders (R,G,B)
-- Reads input from a single (proprietary) input file,
-- built from 24b bitmap image data
-- Produces 3 output files, one per color component, 
-- containing compressed data
-- Compression rate is calculated and displayed at the end

entity enc_tb is
	generic(
		constant IN_WIDTH  : integer := 8;
		constant OUT_WIDTH : integer := 16;
		-- Half the clock period. 
		-- The frequency will be 1/(2*PERIOD) = 100 MHz
		constant PERIOD    : time    := 50 ns;
		
		-- file name affix ( Note: 	the simulation expects the file
		--                   	 	to have 'vhdl' as an suffix)
		constant fil       : string  := "input_file"; 
		
		-- Folder where input file is located
		constant path      : string  := "..\source\";
		
		-- Used to scale the compression calculation
		-- Should be the number of pixels in the input
		constant size      : real    := 1920.0 * 1080.0 
	);                                  
end entity enc_tb;

architecture RTL of enc_tb is
	signal clk      : std_logic;
	signal rst      : std_logic;
	signal inputR   : std_logic_vector(IN_WIDTH - 1 downto 0);
	signal inputG   : std_logic_vector(IN_WIDTH - 1 downto 0);
	signal inputB   : std_logic_vector(IN_WIDTH - 1 downto 0);
	signal newFrame : std_logic;
	signal rdyR     : std_logic;
	signal rdyG     : std_logic;
	signal rdyB     : std_logic;

	signal outputR : std_logic_vector(OUT_WIDTH - 1 downto 0);
	signal outputG : std_logic_vector(OUT_WIDTH - 1 downto 0);
	signal outputB : std_logic_vector(OUT_WIDTH - 1 downto 0);

	signal writeR : std_logic;
	signal writeG : std_logic;
	signal writeB : std_logic;

	-- Enc -> wbuf connect
	signal encRwbufR : std_logic_vector(GOL_MAX_LEN - 1 downto 0);
	signal encGwbufG : std_logic_vector(GOL_MAX_LEN - 1 downto 0);
	signal encBwbufB : std_logic_vector(GOL_MAX_LEN - 1 downto 0);

	signal sizeOutR : std_logic_vector(3 downto 0);
	signal sizeOutG : std_logic_vector(3 downto 0);
	signal sizeOutB : std_logic_vector(3 downto 0);

	signal wbufEmptyR : std_logic;
	signal wbufEmptyG : std_logic;
	signal wbufEmptyB : std_logic;

	signal wbufEn : std_logic;

	shared variable numPixelsIn         : integer := 0;
	shared variable bitsR, bitsG, bitsB : real    := 0.0;
	shared variable bR, bG, bB          : integer := 0;

	signal qR, qG, qB : std_logic_vector(7 downto 0);

	signal delayWbufEn : std_logic := '0';
	signal writeQ      : std_logic := '0';

begin

	-- ENCODER INIT
	encR : entity work.enc_top
		port map(clk       => clk,
			     rst       => rst,
			     data_in   => inputR,
			     new_frame => newFrame,
			     ready     => rdyR,
			     data_out  => encRwbufR,
			     size_out  => sizeOutR
		);
	encG : entity work.enc_top
		port map(clk       => clk,
			     rst       => rst,
			     data_in   => inputG,
			     new_frame => newFrame,
			     ready     => rdyG,
			     data_out  => encGwbufG,
			     size_out  => sizeOutG
		);

	encB : entity work.enc_top
		port map(clk       => clk,
			     rst       => rst,
			     data_in   => inputB,
			     new_frame => newFrame,
			     ready     => rdyB,
			     data_out  => encBwbufB,
			     size_out  => sizeOutB
		);

	-- GOLOMB PACKER (WRITE BUFFER) INIT
	wbufR : entity work.wbuf
		port map(clk         => clk,
			     rst         => rst,
			     in_wbuf     => encRwbufR,
			     word_length => sizeOutR,
			     wbuf_enable => wbufEn,
			     out_wbuf    => outputR,
			     wbuf_empty  => wbufEmptyR,
			     wbuf_write  => writeR);
	wbufG : entity work.wbuf
		port map(clk         => clk,
			     rst         => rst,
			     in_wbuf     => encGwbufG,
			     word_length => sizeOutG,
			     wbuf_enable => wbufEn,
			     out_wbuf    => outputG,
			     wbuf_empty  => wbufEmptyG,
			     wbuf_write  => writeG);
	wbufB : entity work.wbuf
		port map(clk         => clk,
			     rst         => rst,
			     in_wbuf     => encBwbufB,
			     word_length => sizeOutB,
			     wbuf_enable => wbufEn,
			     out_wbuf    => outputB,
			     wbuf_empty  => wbufEmptyB,
			     wbuf_write  => writeB);

	clock_generator : process
	begin
		clk <= '1';
		wait for PERIOD / 2;
		clk <= '0';
		wait for PERIOD / 2;
	end process;

	--/ ---------------------------------
	--/ CONNECT SIGNALS ACROSS DIFFERENT 
	--/ MODULES
	--/ Used to probe entity signals that aren't 
	--/ connected directly to ports 
	drive_sig_process : process
	begin
		init_signal_driver("/enc_tb/encR/Q", "qR");
		init_signal_driver("/enc_tb/encG/Q", "qG");
		init_signal_driver("/enc_tb/encB/Q", "qB");
		wait;
	end process drive_sig_process;
	-- ---------------------------------

	--/ ---------------------------------
	--/ CONTROLLING WHEN WBUF IS ENABLED 
	--/ AND WHEN TO START WRITING Q's
	--/ The Golomb packer is enabled 1 cycle before the output is valid
	--/ and stopped when the last valid output has passed.
	--/ writeQ is only used to probe the Q signals within the encoder 
	--/ entities for debugging convenience.
	wbufEnable : process(clk) is
		variable count : std_logic_vector(3 downto 0) := (others => '0');
	begin
		if (rising_edge(clk)) then
			if (delayWbufEn = '1') then
				if (count > "0101") then
					wbufEn <= '1';
					writeQ <= '1';
				elsif (count > "0011") then
					writeQ <= '1';
					wbufEn <= '0';
					count  := count + "1";
				else
					wbufEn <= '0';
					writeQ <= '0';
					count  := count + "1";
				end if;
			else
				if (count > "0101") then
					wbufEn <= '1';
					writeQ <= '1';
					count  := count - "1";
				elsif (count > "0001") then
					wbufEn <= '1';
					writeQ <= '0';
					count  := count - "1";
				else
					wbufEn <= '0';
					writeQ <= '0';
				end if;
			end if;
		end if;
	end process;
	-- ---------------------------------
	--/ PROCESS CONTROLLING INPUT
	--/ Input is read from the file defined with the constants
	--/ Loops new input every clock cycle until the input file
	--/ is empty.
	--/ Disable signal is triggered, and statistics are calculated

	inputctrl : process
		file inn : text is in path & fil & "vhdl";
		variable buf_in, buf_nSymb : line;
		variable inV               : std_logic_vector(23 downto 0) 
																											:= (others => '0');
		variable cR, cG, cB, cT    : real;
		variable overhead          : integer;
	begin
		rst <= '1';
		wait for PERIOD;
		inputR <= (others => '0');
		inputG <= (others => '0');
		inputB <= (others => '0');
		rst    <= '0';

		wait for PERIOD;
		newFrame <= '1';
		while (rdyR = '0' or rdyG = '0' or rdyB = '0') loop
			wait for PERIOD;
		end loop;
		wait for PERIOD;
		wait for 5 ns;
		newFrame <= '0';
		WAIT FOR 45 ns;
		wait for 9 * PERIOD;
		delayWbufEn <= '1';

		while not ENDFILE(inn) loop
			-- Get line from input file
			READLINE(inn, buf_in);
			-- Save value to variable
			HREAD(buf_in, inV);

			-- Spectral decorrelation of the inputs.
			inputR <= inV(23 downto 16) - inV(15 downto 8);
			inputG <= inV(15 downto 8);
			inputB <= inV(7 downto 0) - inV(15 downto 8);

			-- TESTING UTEN R-G,G,B-G DEBUG!!!
			--		inputR <= inV(23 downto 16);
			--		inputG <= inV(15 downto 8);
			--		inputB <= inV(7 downto 0);

			-- Count pixels
			numPixelsIn := numPixelsIn + 1;
			wait for PERIOD;
		end loop;
		inputR <= (others => '1');
		inputG <= (others => '1');
		inputB <= (others => '1');
		wait for PERIOD;
		delaywbufEn <= '0';

		wait for 10 * PERIOD;

		-- -------------------------------------
		-- END SIMULATION : CALCULATE STATISTICS
		-- Compression per component and total
		-- Overhead added for different packet 
		-- sizes. (large packets, not golomb packets) 
		-- -------------------------------------
		cR       := bitsR / real'(size);
		cG       := bitsG / real'(size);
		cB       := bitsB / real'(size);
		overhead := (bR + bG + bB) * 2 / 16 / 8000;
		cT       := (bitsR + bitsG + bitsB) / real'(size * 3.0);
		write(OUTPUT, "Ending simulation." & LF);
		write(OUTPUT, "Simulation statistics for " & fil & ":" & LF);
		write(OUTPUT, " ----------------- " & LF);
		write(OUTPUT, "R   : " & real'image(cR) & LF);
		write(OUTPUT, "G   : " & real'image(cG) & LF);
		write(OUTPUT, "B   : " & real'image(cB) & LF);
		write(OUTPUT, " - - - - - - - - - " & LF);
		write(OUTPUT, "Tot : " & real'image(cT) & LF);
		write(OUTPUT, " - - - - - - - - - " & LF);
		write(OUTPUT, "Head 16 : " & natural'image(overhead)&" KB" & LF);
		overhead := overhead / 2;
		write(OUTPUT, "Head 32 : " & natural'image(overhead)&" KB" & LF);
		overhead := overhead / 2;
		write(OUTPUT, "Head 64 : " & natural'image(overhead)&" KB" & LF);
		overhead := overhead / 2;
		write(OUTPUT, "Head 128: " & natural'image(overhead)&" KB" & LF);
		overhead := overhead / 2;
		write(OUTPUT, "Head 256: " & natural'image(overhead)&" KB" & LF);

		write(OUTPUT, " ----------------- " & LF);
		write(OUTPUT, "Simulation complete.");
		report "" severity failure;
	-- -------------------------------------
	end process;

	-- ---------------------------------
	--/ PROCESS CONTROLLING OUTPUT
	--/ Output is written every time the write signals are high
	--/ 3 output files, 1 per color component.
	--/ The total number of bits outputted are counted for use
	--/ in compression calculation.
	--/ Additional information is written for information (packet order) 
	--/ and debugging (probed Q from within encoder entities) purposes.

	writectrl : process(clk)
		file outR : text open WRITE_MODE 
			is path & fil & "\" & fil & "_outR";
		file outG : text open WRITE_MODE 
			is path & fil & "\" & fil & "_outG";
		file outB : text open WRITE_MODE 
			is path & fil & "\" & fil & "_outB";

		file wrorder : text open WRITE_MODE 
			is path & fil & "\" & fil & "_wrorder";

		file qoutR : text open WRITE_MODE 
			is path & fil & "\" & fil & "_qR";
		file qoutG : text open WRITE_MODE 
			is path & fil & "\" & fil & "_qG";
		file qoutB : text open WRITE_MODE 
			is path & fil & "\" & fil & "_qB";

		variable buf_outR, buf_outG, buf_outB, buf_qR, 
			buf_qG, buf_qB, buf_wOrdr : line;

	begin
		if (rising_edge(clk)) then
			if (writeR = '1') then
				WRITE(buf_outR, outputR(15 downto 0), RIGHT, OUT_WIDTH / 2);
				WRITELINE(outR, buf_outR);
				bitsR := bitsR + 16.0;
				bR    := bR + 16;

				WRITE(buf_wOrdr, 1);
				WRITELINE(wrorder, buf_wOrdr);
			end if;
			if (writeG = '1') then
				WRITE(buf_outG, outputG(15 downto 0), RIGHT, OUT_WIDTH / 2);
				WRITELINE(outG, buf_outG);
				bitsG := bitsG + 16.0;
				bG    := bG + 16;

				WRITE(buf_wOrdr, 2);
				WRITELINE(wrorder, buf_wOrdr);
			end if;
			if (writeB = '1') then
				WRITE(buf_outB, outputB(15 downto 0), RIGHT, OUT_WIDTH / 2);
				WRITELINE(outB, buf_outB);
				bitsB := bitsB + 16.0;
				bB    := bB + 16;

				WRITE(buf_wOrdr, 3);
				WRITELINE(wrorder, buf_wOrdr);
			end if;

			if (writeQ = '1') then
				HWRITE(buf_qR, qR, RIGHT, 2);
				WRITELINE(qoutR, buf_qR);

				HWRITE(buf_qG, qG, RIGHT, 2);
				WRITELINE(qoutG, buf_qG);

				HWRITE(buf_qB, qB, RIGHT, 2);
				WRITELINE(qoutB, buf_qB);
			end if;

		end if;
	end process;
end architecture RTL;
